@@ -1,5 +1,5 @@
 /*
-  Copyright 2007-2011 David Robillard <http://drobilla.net>
+  Copyright 2007-2012 David Robillard <http://drobilla.net>
 
   Permission to use, copy, modify, and/or distribute this software for any
   purpose with or without fee is hereby granted, provided that the above
@@ -14,20 +14,18 @@
   OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
 
-#define _XOPEN_SOURCE 500
-
 #include <assert.h>
 #include <stdbool.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
 
-#include "suil-config.h"
-#include "suil_internal.h"
+#include "./suil_config.h"
+#include "./suil_internal.h"
 
-#define NS_UI       "http://lv2plug.in/ns/extensions/ui#"
-#define GTK2_UI_URI NS_UI "GtkUI"
-#define QT4_UI_URI  NS_UI "Qt4UI"
+#define GTK2_UI_URI LV2_UI__GtkUI
+#define QT4_UI_URI  LV2_UI__Qt4UI
+#define X11_UI_URI  LV2_UI__X11UI
 
 SUIL_API
 unsigned
@@ -44,23 +42,23 @@ suil_ui_supported(const char* container_type_uri,
 	} else if ((!strcmp(container_type_uri, GTK2_UI_URI)
 	            && !strcmp(ui_type_uri, QT4_UI_URI))
 	           || (!strcmp(container_type_uri, QT4_UI_URI)
-	               && !strcmp(ui_type_uri, GTK2_UI_URI))) {
+	               && !strcmp(ui_type_uri, GTK2_UI_URI))
+	           || (!strcmp(container_type_uri, GTK2_UI_URI)
+	               && !strcmp(ui_type_uri, X11_UI_URI))
+	           || (!strcmp(container_type_uri, QT4_UI_URI)
+	               && !strcmp(ui_type_uri, X11_UI_URI))) {
 		return SUIL_WRAPPING_EMBEDDED;
 	} else {
 		return SUIL_WRAPPING_UNSUPPORTED;
 	}
 }
 
-struct _SuilModule {
-	SuilWrapInitFunc init;
-	SuilWrapFunc     wrap;
-};
-
-typedef struct _SuilModule* SuilModule;
-
-static SuilModule
-get_wrap_module(const char* container_type_uri,
-                const char* ui_type_uri)
+static SuilWrapper*
+open_wrapper(SuilHost*      host,
+             const char*    container_type_uri,
+             const char*    ui_type_uri,
+             LV2_Feature*** features,
+             unsigned       n_features)
 {
 	if (!strcmp(container_type_uri, ui_type_uri)) {
 		return NULL;
@@ -71,8 +69,14 @@ get_wrap_module(const char* container_type_uri,
 	    && !strcmp(ui_type_uri, GTK2_UI_URI)) {
 		module_name = "libsuil_gtk2_in_qt4";
 	} else if (!strcmp(container_type_uri, GTK2_UI_URI)
-	    && !strcmp(ui_type_uri, QT4_UI_URI)) {
+	           && !strcmp(ui_type_uri, QT4_UI_URI)) {
 		module_name = "libsuil_qt4_in_gtk2";
+	} else if (!strcmp(container_type_uri, GTK2_UI_URI)
+	           && !strcmp(ui_type_uri, X11_UI_URI)) {
+		module_name = "libsuil_x11_in_gtk2";
+	} else if (!strcmp(container_type_uri, QT4_UI_URI)
+	           && !strcmp(ui_type_uri, X11_UI_URI)) {
+		module_name = "libsuil_x11_in_qt4";
 	}
 
 	if (!module_name) {
@@ -98,20 +102,29 @@ get_wrap_module(const char* container_type_uri,
 		return NULL;
 	}
 
-	SuilModule module = (SuilModule)malloc(sizeof(struct _SuilModule));
-	module->init = (SuilWrapInitFunc)suil_dlfunc(lib, "suil_wrap_init");
-	module->wrap = (SuilWrapFunc)suil_dlfunc(lib, "suil_wrap");
+	SuilWrapperNewFunc wrapper_new = (SuilWrapperNewFunc)suil_dlfunc(
+		lib, "suil_wrapper_new");
 
-	if (!module->init || !module->wrap) {
+	SuilWrapper* wrapper = wrapper_new
+		? wrapper_new(host,
+		              container_type_uri,
+		              ui_type_uri,
+		              features,
+		              n_features)
+		: NULL;
+
+	if (!wrapper) {
 		SUIL_ERRORF("Corrupt module %s\n", path);
+		dlclose(lib);
 		free(path);
-		free(module);
 		return NULL;
 	}
 
 	free(path);
 
-	return module;
+	wrapper->lib = lib;
+
+	return wrapper;
 }
 
 SUIL_API
@@ -161,52 +174,67 @@ suil_instance_new(SuilHost*                 host,
 		return NULL;
 	}
 
-	// Use empty local features array if necessary
-	const LV2_Feature* local_features[1];
-	local_features[0] = NULL;
-	if (!features) {
-		features = (const LV2_Feature* const*)&local_features;
+	// Create SuilInstance
+	SuilInstance* instance = calloc(1, sizeof(struct SuilInstanceImpl));
+	instance->lib_handle = lib;
+	instance->descriptor = descriptor;
+
+	// Make UI features array
+	instance->features = (LV2_Feature**)malloc(sizeof(LV2_Feature**));
+	instance->features[0] = NULL;
+
+	// Copy user provided features
+	unsigned n_features = 0;
+	for (; features && features[n_features]; ++n_features) {
+		const LV2_Feature* f = features[n_features];
+		suil_add_feature(&instance->features, n_features, f->URI, f->data);
 	}
 
-	SuilModule module = get_wrap_module(container_type_uri, ui_type_uri);
-	if (module) {
-		module->init(host, container_type_uri, ui_type_uri, features);
+	// Add additional features implemented by SuilHost functions
+	if (host->index_func) {
+		instance->port_map.handle     = controller;
+		instance->port_map.port_index = host->index_func;
+		suil_add_feature(&instance->features, n_features++,
+		                 LV2_UI__portMap, &instance->port_map);
 	}
+	if (host->subscribe_func && host->unsubscribe_func) {
+		instance->port_subscribe.handle      = controller;
+		instance->port_subscribe.subscribe   = host->subscribe_func;
+		instance->port_subscribe.unsubscribe = host->unsubscribe_func;
+		suil_add_feature(&instance->features, n_features++,
+		                 LV2_UI__portSubscribe, &instance->port_subscribe);
+	}
+	if (host->touch_func) {
+		instance->touch.handle = controller;
+		instance->touch.touch  = host->touch_func;
+		suil_add_feature(&instance->features, n_features++,
+		                 LV2_UI__touch, &instance->touch);
+	}
+
+	// Open wrapper (this may add additional features)
+	instance->wrapper = open_wrapper(
+		host, container_type_uri, ui_type_uri, &instance->features, n_features);
 
 	// Instantiate UI
-	SuilInstance* instance = malloc(sizeof(struct SuilInstanceImpl));
-	instance->lib_handle  = lib;
-	instance->descriptor  = descriptor;
-	instance->host_widget = NULL;
-	instance->ui_widget   = NULL;
-	instance->handle      = descriptor->instantiate(
+	instance->handle = descriptor->instantiate(
 		descriptor,
 		plugin_uri,
 		ui_bundle_path,
 		host->write_func,
 		controller,
 		&instance->ui_widget,
-		features);
+		(const LV2_Feature* const*)instance->features);
 
-	// Failed to find or instantiate UI
+	// Failed to instantiate UI
 	if (!instance || !instance->handle) {
 		SUIL_ERRORF("Failed to instantiate UI <%s> in %s\n",
 		            ui_uri, ui_binary_path);
-		free(instance);
-		dlclose(lib);
-		return NULL;
-	}
-
-	// Got a handle, but failed to create a widget (buggy UI)
-	if (!instance->ui_widget) {
-		SUIL_ERRORF("Widget creation failed for UI <%s> in %s\n",
-		            ui_uri, ui_binary_path);
 		suil_instance_free(instance);
 		return NULL;
 	}
 
-	if (module) {
-		if (module->wrap(container_type_uri, ui_type_uri, instance)) {
+	if (instance->wrapper) {
+		if (instance->wrapper->wrap(instance->wrapper, instance)) {
 			SUIL_ERRORF("Failed to wrap UI <%s> in type <%s>\n",
 			            ui_uri, container_type_uri);
 			suil_instance_free(instance);
@@ -224,7 +252,18 @@ void
 suil_instance_free(SuilInstance* instance)
 {
 	if (instance) {
-		instance->descriptor->cleanup(instance->handle);
+		for (unsigned i = 0; instance->features[i]; ++i) {
+			free(instance->features[i]);
+		}
+		free(instance->features);
+
+		if (instance->wrapper) {
+			instance->wrapper->free(instance->wrapper);
+			dlclose(instance->wrapper->lib);
+		}
+		if (instance->handle) {
+			instance->descriptor->cleanup(instance->handle);
+		}
 		dlclose(instance->lib_handle);
 		free(instance);
 	}
